; MIDI on MPU-401 driver for SCI0
; plays the GM channels on new SCI0 games
; assembles with NASM (see http://sourceforge.net/projects/nasm/):
;    nasm mpumidi.asm -o mpumidi.drv -f bin

; Copyright (c) 1999, 2000, 2002 by Ravi Iyengar [ravi.i@softhome.net]
;
; This library is free software; you can redistribute it and/or
; modify it under the terms of the GNU Lesser General Public
; License as published by the Free Software Foundation; either
; version 2.1 of the License, or (at your option) any later version.
;
; This library is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
; Lesser General Public License for more details.
;
; You should have received a copy of the GNU Lesser General Public
; License along with this library; if not, write to the Free Software
; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


[BITS 16]

; Drivers are flatform binaries that get loaded at the start of
; a segment. When a driver function needs to be executed, the
; interpreter does a far call to offset 0. The first thing must
; be a jump to a DriverInterface function because the interpreter
; expects certain data starting at offset 3, and there isn't
; space to fit in anything else.

jmp near DriverInterface

; identifiers for a Sierra sound driver
db 0                   ; uncertain
dd 0x87654321          ; Sierra driver id
db 1                   ; sound driver id (0=display, 4=keyboard)

; a string identifying the driver
db 10, 'mpumididrv'

; a string with a description of the device
db 24, 'MPU MIDI (GM) BETA 0.5.1'



;--------------------------------------------------------------------------
; misc macros and defines

; device information returned by GetDeviceInfo
%define  INFO_PATCH  0xFFFF  ; patch resource number (0xFFFF means none)
%define  INFO_POLY   32      ; maximum polyphony

; the play flag for this device (used in the sound header)
%define  PLAY_FLAG   0x80

; playstate constants
%define  STOPPED     0
%define  PLAYING     1
%define  SEEKING     2

; offsets in the sound info structure
%define  SND_RESPTR  8       ; heap pointer to far pointer to resource data
%define  SND_POS     12      ; current play offset in sound resource
%define  SND_STATE   16      ; 1 = valid, 3 = invalid   (possibly others)
%define  SND_SIGNAL  22      ; the Sound object's signal property
%define  SND_VOLUME  24      ; global sound volume



;--------------------------------------------------------------------------
; data area
; Sierra's drivers put all their data space in the beginning, though
;    there's no reason why it has to be that way.

ptrparam:      dw 0          ; saves the original si parameter to DriverInterface
playstate:     db 0          ; 0 = not playing sound, 1 = playing sound, 2 = seeking sound
fadeticks:     db 0          ; number of ticks to next volume decrement, 0 = not fading
fadevolume:    db 0          ; the volume setting during fades
playflags:     dw 0          ; specifies whether or not each channel is used by this device
waitcount:     dw 0          ; number of ticks left before next event
sndresptr:     dw 0          ; pointer to sound resource
looppoint:     dw 0          ; sound loop position
cumcue:        dw 0          ; the cumulative cue
status:        db 0          ; saves the last status for running status mode, holds the current status
pssndreset:    db 0          ; boolean describing whether PauseSound should reset the sound
seekto:        dw 0          ; the end point for SeekSound
seekcue:       dw 0          ; the desired sound cue at the end of SeekSound
globalvol:     dw 60         ; global volume on a scale of 0 to 63


;--------------------------------------------------------------------------
; export area
; the interpreter interacts with the driver through these functions


; Export Lookup Table
; used by DriverInterface to call exported functions by number
; Exports must be even numbers because offsets are 2 bytes each.
;    Later interpreters used odd numbers for export functions as
;    well and multiplied by 2 to get the offset in this table.

ExportTable:
   Export0:    dw GetDeviceInfo
   Export2:    dw InitDevice
   Export4:    dw ShutdownDevice
   Export6:    dw LoadSound
   Export8:    dw DoSoundEvents
   Export10:   dw SetVolume
   Export12:   dw FadeOut
   Export14:   dw StopSound
   Export16:   dw PauseSound
   Export18:   dw SeekSound


; DriverInterface
; Saves registers and calls export
; Entry: bp = export number
;        ds = heap segment
;        si = heap pointer where applicable
; Exit: ax, cx = return values where applicable
;       data at ds:si modified where applicable

DriverInterface:
   pushf
   push dx
   push bx
   push sp
   push bp
   push si
   push di
   push es
   push ds
   mov [cs:ptrparam], si
   call near word [cs:ExportTable+bp]
   pop ds
   pop es
   pop di
   pop si
   pop bp
   pop sp
   pop bx
   pop dx
   popf
   retf


; GetDeviceInfo
; Called at startup to determine max poly and which patch
;    resource to load.
; Entry: nothing
; Exit: ax = patch number (0xFFFF if no patch)
;       cx = maximum polyphony

GetDeviceInfo:
   mov ax, INFO_PATCH
   mov cx, INFO_POLY
   retn


; InitDevice
; Called to load the patch if applicable and initialize the device
; Entry: ds:si points to patch resource structure
; Exit: cs:ax points to the area of the driver segment which can be overwritten
;       cx = 0

InitDevice:
   call mpu_detect
   cmp ah, 0xFF
   jne have_mpu

   mov ax, 0xFFFF
   xor cx, cx
   retn

   have_mpu:
   call mpu_init
   mov ax, temp_area
   xor cx, cx
   retn


; ShutdownDevice
; Called to close the device when the interpreter exits. There
;    is no guarantee that StopSound will have been called on a
;    currently playing sound.
; Entry: ds:si points to sound resource structure
; Exit: nothing

ShutdownDevice:
   call mpu_silence
   call mpu_exit
   retn


; LoadSound
; Called before a sound starts playing so the driver has a chance
;    to load the header and perform initialization work.
; Entry: ds:si points to sound resource structure
; Exit: ax = 3 on unrecognized header byte, 1 on success

LoadSound:
   mov word [cs:playflags], 0x8000 ; channel 15 (the control channel) should always be enabled

   mov bx, [si+SND_RESPTR]
   mov es, [bx+2]
   mov di, [bx]              ; es:di points to sound data

   xor dx, dx                ; dx will be added to di to compensate for missing channel 15 info
   mov cx, 16
   mov al, byte [es:di]      ; check for the presence of a PCM sample
   inc di
   cmp al, 0
   je load_header
   cmp al, 2
   je fifteen_chan

   mov ax, 3                 ; unplayable / invalid resource
   mov word [si+SND_STATE], ax
   retn

   fifteen_chan:
   dec cx                     ; if there is a PCM sample, the header only has info for 15 channels
   mov dx, 2                  ; compensates for missing channel 15 info

   load_header:
   mov bx, 1

   load_header_loop:
   mov al, byte [es:di]      ; al = initial number of voices on this channel
   inc di
   mov ah, byte [es:di]      ; ah = play flag for this channel
   inc di
   test ah, PLAY_FLAG        ; check for playflag
   jz continue_load

   or word [cs:playflags], bx ; enable channel

   continue_load:
   shl bx, 1
   loop load_header_loop

   add di, dx                 ; dx = 2 if there was only init data for 15 channels

   call getwaitcount

   mov ax, 1
   mov word [si+SND_STATE], ax  ; success loading sound
   mov word [si+SND_POS], di    ; set the position to the start of sound data
   mov word [si+SND_SIGNAL], 0  ; set the signal to 0
   mov word [cs:cumcue], 127    ; initialize the cumulative cue
   mov word [cs:looppoint], 33  ; initialize the loop point to the top
   mov word [cs:pssndreset], 0  ; PauseSound does not reset by default
   mov byte [cs:fadeticks], 0   ; there is no fade in progress
   mov byte [cs:playstate], PLAYING ; we're playing this sound now
   
   mov cx, word [si+SND_VOLUME]
   shl cx, 2                 ; we use a range of 0-63 for internal global volume
   mov word [cs:globalvol], cx

   call mpu_reset

   retn


; DoSoundEvents
; Called each MIDI tick (60 times per second) to actually play the
;    sound resource
; Entry: ds:si points to sound resource structure
; Exit: nothing

StatusCallTable:             ; call table for handling status messages
   dw StopNote
   dw PlayNote
   dw KeyPressure
   dw ControlChange
   dw PatchChange
   dw ChanPressure
   dw PitchWheel

ParamCount: db 2, 2, 2, 2, 1, 1, 2 ; number of parameters used by each status

DoSoundEvents:
   mov bx, [si+SND_RESPTR]
   mov es, [bx+2]
   mov di, [bx]              ; es:di points to sound resource
   mov word [cs:sndresptr], di
   add di, word [si+SND_POS] ; get the current sound resource position

   call fader
   cmp byte [cs:playstate], STOPPED
   jne do_frame
   retn                      ; the sound isn't playing anymore, so quit

   do_frame:                 ; the loop to do all events in this frame (all events with delta time = 0)

   cmp word [cs:waitcount], 0 ; check if there's a new event
   je next_event
   dec word [cs:waitcount]
   jmp end_frame

   next_event:
   xor cx, cx
   mov cl, byte [es:di]
   inc di
   test cl, 0x80             ; check for running status mode
   jnz check_status
   mov cl, byte [cs:status]  ; repeat the last status
   test cl, 0x80             ; sanity check
   jz do_frame               ;   - the first event uses running status, which is insane for this driver
   dec di                    ; rewind di for re-reading the last byte as a parameter

   check_status:
   mov dl, cl                ; dl = full status byte (status and channel)
   mov byte [cs:status], cl
   mov al, cl
   shr al, 4                 ; al = status
   and cl, 0x0F              ; cl = channel

   cmp al, 0x0F
   je sys_message            ; check for a system message before we check the channel

   mov bx, 1
   shl bx, cl
   test word [cs:playflags], bx ; test to see if this channel is enabled
   jnz do_status

   mov bl, al                ; al is still the status
   sub bl, 8
   xor bh, bh
   mov bl, [cs:ParamCount+bx]
   add di, bx
   call getwaitcount
   jmp do_frame              ; this channel is to be ignored, so skip this status

   do_status:
   mov bl, al                ; al is still the status
   sub bl, 8
   xor bh, bh
   shl bx, 1
   call word [cs:StatusCallTable+bx] ; call the handler for this particular status
   call getwaitcount
   jmp do_frame

   sys_message:
   ; FIXME: We assume that this is 0xFC, but we should really check
   mov word [si+SND_SIGNAL], 0xFFFF ; the sound is done - set the signal appropriately
   xor bx, bx
   mov [cs:waitcount], bx
   mov bx, word [si+SND_RESPTR]
   mov di, word [bx]
   add di, word [cs:looppoint] ; set es:di to appropriate loop point

   end_frame:
   sub di, word [cs:sndresptr]
   mov word [si+SND_POS], di ; update the sound position
   retn


; SetVolume
; Called when the user changes the global sound volume (Ctrl-V in the game)
; Entry: ds:si points to sound resource structure
; Exit: nothing

SetVolume:
   mov cx, [si+SND_VOLUME]   ; cx = volume level (0-15)
   shl cx, 2                 ; we use a range of 0-63 for internal global volume
   mov [cs:globalvol], cx
   retn


; FadeOut
; Called to slowly fade and stop the currently playing sound
; Entry: ds:si points to sound resource structure
; Exit: nothing

FadeOut:
   mov cx, [si+SND_VOLUME]   ; cx = volume level (0-15)
   shl cx, 2                 ; we use a range of 0-63 for internal global volume
   test cx, cx
   jz f_stop                 ; just stop the sound now if the volume starts at 0

   mov [cs:fadevolume], cl
   mov byte [cs:fadeticks], 5
   retn

   f_stop:
   mov word [si+SND_SIGNAL], 0xFFFF ; the sound is done - set the signal appropriately
   call mpu_silence
   mov byte [cs:playstate], STOPPED
   retn


; StopSound
; Stops playback by turning off sound output. Nothing needs to be done to
;   prevent further MIDI events from occurring because the interpreter will
;   simply stop calling DoSoundEvents.
; Entry: ds:si points to sound resource structure
; Exit: the sound's signal property is set to -1

StopSound:
   call mpu_silence
   mov word [si+SND_SIGNAL], 0xFFFF
   mov word [cs:playstate], STOPPED
   retn


; PauseSound
; Pauses playback by turning off sound output. Nothing needs to be done to
;   prevent further MIDI events from occurring because the interpreter will
;   simply stop calling DoSoundEvents until playback resumes. If CC4C has been
;   set, the sound position should be reset. Otherwise it should be left alone.
; Entry: ds:si points to sound resource structure
; Exit: nothing

PauseSound:
   call mpu_silence

   cmp byte [cs:pssndreset], 0
   je exitpausesound

   mov ax, word [cs:looppoint] ; reset sound position to loop point
   mov word [si+SND_POS], ax

   mov bx, [si+SND_RESPTR]
   mov es, [bx+2]
   mov di, [bx]              ; es:di points to sound resource
   add di, ax                ; move forward to the loop position
   call getwaitcount         ; get a new waitcount for when the sound resumes

   exitpausesound:
   retn


; SeekSound
; Called when a sound needs to be quickly set to a certain point (e.g. on
;    restore, when one sound interrupts another, and so on)
; Entry: ds:si points to sound resource structure
; Exit: nothing

SeekSound:
   mov ax, word [si+SND_POS] ; this is where we need to seek to
   mov word [cs:seekto], ax
   mov ax, word [si+SND_SIGNAL] ; this is the what the cue should be
   call LoadSound
   mov byte [cs:playstate], SEEKING
   cmp ax, 1
   je do_seek
   retn

   do_seek:
   call DoSoundEvents
   mov bx, word [si+SND_POS]
   cmp bx, word [cs:seekto]
   jb do_seek                ; make sure we're at the right sound position
   mov bx, word [si+SND_SIGNAL]
   cmp bx, word [cs:seekcue]
   jne do_seek               ; make sure we're at the right cue position

   mov byte [cs:playstate], PLAYING
   retn



;--------------------------------------------------------------------------
; status functions

; on entry for each of these functions:
;   al = status
;   cl = channel
;   dl = full status byte (status and channel)
;   es:di points to parameters for the status
; on exit:
;   es:di points to the delta time until the next sound element


; two parameter status messages that get passed through to the device
KeyPressure:
PitchWheel:
   mov al, dl
   call mpu_output
   mov al, byte [es:di]
   call mpu_output
   mov al, byte [es:di+1]
   call mpu_output
   add di, 2
   ret


StopNote:
   cmp byte [cs:playstate], PLAYING
   je stop_note

   add di, 2                 ; we're just seeking, don't play anything
   ret

   stop_note:
   mov al, dl
   call mpu_output
   mov al, byte [es:di]
   call mpu_output
   mov al, byte [es:di+1]
   call mpu_output
   add di, 2
   ret


PlayNote:
   cmp byte [cs:playstate], PLAYING
   je play_note

   add di, 2                 ; we're just seeking, don't play anything
   ret

   play_note:
   mov al, dl
   call mpu_output
   mov al, byte [es:di]
   call mpu_output
   mov al, byte [es:di+1]
   cbw
   mul word [cs:globalvol]   ; multiply by [cs:globalvol]/64 to scale volume
   mov bx, 64
   div bx
   call mpu_output
   add di, 2
   ret


ControlChange:
   mov bh, byte [es:di]      ; bh = control number
   mov bl, byte [es:di+1]    ; bl = control value
   add di, 2
   cmp bh, 0x60
   jne not_CC60h

   xor bh, bh
   add bx, word [cs:cumcue]
   mov word [si+SND_SIGNAL], bx ; set the new signal
   mov word [cs:cumcue], bx  ; update our cumulative cue variable
   ret

   not_CC60h:
   cmp bh, 0x4C
   jne not_CC4Ch

   mov byte [cs:pssndreset], bl ; set the PauseSound reset control value
   ret

   not_CC4Ch:
   mov al, dl
   call mpu_output
   mov al, bh
   call mpu_output
   mov al, bl
   call mpu_output
   ret


PatchChange:
   mov bl, byte [es:di]      ; bl = patch
   inc di
   cmp cl, 15
   jne normal_patch_change

   cmp bl, 127
   je set_loop_point

   mov byte [si+SND_SIGNAL], bl ; this is a cue - set the signal property
   ret

   set_loop_point:           ; set the sound loop point
   push di
   sub di, word [cs:sndresptr]
   sub di, 2
   mov word [cs:looppoint], di
   pop di
   ret

   normal_patch_change:      ; this is a simple patch change
   mov al, dl
   call mpu_output
   mov al, bl
   call mpu_output
   ret


ChanPressure:
   mov al, dl
   call mpu_output
   mov al, [es:di]
   call mpu_output
   add di, 1
   ret



;----------------------------------------------------------------------------
; MPU functions


; wait_for_mpu
; wait for the specified bit to clear on the specified port
; in: ah = flag
; out: al = 0 if cleared, -1 if timed out

wait_for_mpu:
   push cx
   push dx
   mov dx, 0x331
   mov cx, 0x7FFF            ; timeout
   wfm_waiter:
   in al, dx
   and al, ah
   jz wfm_success
   loop wfm_waiter

   mov al, 0xFF              ; fail
   pop dx
   pop cx
   ret

   wfm_success:              ; succes
   pop dx
   pop cx
   ret


; mpu_output
; output the specified byte on the specified port
; in: al = byte
; out:

mpu_output:
   push ax
   mov ah, 0x40
   call wait_for_mpu
   pop ax
   push dx
   mov dx, 0x330
   out dx, al
   pop dx
   ret


; mpu_input
; input a byte from the specified port
; in:
; out: al = byte, ah = 0 on success, -1 on fail

mpu_input:
   push dx
   mov ah, 0x80
   call wait_for_mpu
   mov ah, al
   mov dx, 0x330
   in al, dx
   pop dx
   ret


; mpu_detect
; check for the existance of the MPU on the specified base port
; in:
; out: ah = 0 on success, -1 on fail, al destroyed

mpu_detect:
   push dx
   mov dx, 0x331
   mov al, 0xFF
   out dx, al
   pop dx
   call mpu_input
   mov ah, 0x40
   call wait_for_mpu
   test al, al
   jnz md_fail

   xor ah, ah                ; success
   ret

   md_fail:
   mov ah, 0xFF              ; fail
   ret


; mpu_init
; initialize the mpu on the specified base port
; in:
; out:

mpu_init:
   push dx
   mov dx, 0x331
   mov al, 0x3F
   out dx, al
   pop dx
   call mpu_input

   mov al, 0xB1
   call mpu_output
   mov al, 121
   call mpu_output
   mov al, 0
   call mpu_output

   ret


; mpu_exit
; close the mpu on the specified base port
; in:
; out:

mpu_exit:
   push dx
   mov dx, 0x331
   mov al, 0xFF
   out dx, al
   pop dx
   ret


; mpu_silence
; silence the mpu
; in:
; out:

mpu_silence:
   push ax
   push cx
   mov cx, 0x0F
   mov ah, 0xB0
   mpu_silence_loop:
   mov al, ah
   or al, cl
   call mpu_output
   mov al, 123
   call mpu_output
   mov al, 0
   call mpu_output
   loop mpu_silence_loop
   pop cx
   pop ax
   ret


; mpu_reset
; reset the mpu for playback of a new sound
; in:
; out:

mpu_reset:
   call mpu_silence
   push ax
   push cx
   mov cx, 0x0F
   mov ah, 0xB0
   mpu_reset_loop:
   mov al, ah
   or al, cl
   call mpu_output
   mov al, 121
   call mpu_output
   mov al, 0
   call mpu_output
   loop mpu_reset_loop
   pop cx
   pop ax
   ret



;--------------------------------------------------------------------------
; misc functions


; getwaitcount
; retrieves the next delta time from the sound resource
; entry: es:di points to the current delta time value
; exit: [cs:waitcount] set, es:di points to the sound element following the delta time

getwaitcount:
   push ax
   push dx
   xor ax, ax
   xor dx, dx

   get_delta_time_loop:
   mov dl, byte [es:di]
   cmp dl, 0xF8
   jne add_dt

   mov dl, 0xF0              ; delta time extension

   add_dt:
   add ax, dx
   inc di
   cmp dl, 0xF0
   je get_delta_time_loop

   mov word [cs:waitcount], ax ; set waitcount
   pop dx
   pop ax
   ret


; fader
; checks for a fade in progress and acts accordingly
; entry: es:di points to the current position in the sound resource
; exit:

fader:
   cmp byte [cs:fadeticks], 0 ; check to see if there's a fade in progress
   je exit_fader

   dec byte [cs:fadeticks]
   cmp byte [cs:fadeticks], 0 ; check if we need to decrease the volume yet
   jne exit_fader

   dec byte [cs:fadevolume]
   cmp byte [cs:fadevolume], 0 ; just stop the sound if we're done fading
   je f_stopsnd

   mov byte [cs:fadeticks], 5
   push cx
   xor cx, cx
   mov cl, byte [cs:fadevolume]
   mov [cs:globalvol], cx
   pop cx
   jmp exit_fader

   f_stopsnd:
   mov byte [cs:playstate], STOPPED
   push si
   mov si, [cs:ptrparam]
   mov word [si+SND_SIGNAL], 0xFFFF ; the sound is done - set the signal appropriately
   call mpu_silence
   pop si

   exit_fader:
   ret



;--------------------------------------------------------------------------
; data / functions that will only be used before or during initialization
temp_area:
